/**
 * EasyBeans
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:$
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.injection.impl;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.CARDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelper;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelperException;
import org.ow2.util.ee.metadata.common.api.ICommonClassMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonMethodMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonFieldMetadata;
import org.ow2.util.ee.metadata.common.api.struct.IJEjbEJB;
import org.ow2.util.ee.metadata.common.api.struct.IJAnnotationResource;
import org.ow2.util.ee.metadata.common.api.interfaces.ISharedMetadata;
import org.ow2.util.ee.metadata.car.api.ICarDeployableMetadata;
import org.ow2.util.ee.metadata.car.api.ICarMethodMetadata;
import org.ow2.util.ee.metadata.car.api.ICarFieldMetadata;
import org.ow2.util.ee.metadata.car.api.ICarClassMetadata;
import org.ow2.util.ee.metadata.car.impl.CarDeployableMetadataFactory;
import org.ow2.util.ee.metadata.car.impl.CarDeployableMetadata;
import org.ow2.easybeans.injection.api.ArchiveInjectionException;
import org.ow2.easybeans.resolver.api.EZBRemoteJNDIResolver;
import org.ow2.easybeans.resolver.api.EZBJNDIData;
import org.ow2.easybeans.asm.Type;
import org.ow2.util.scan.api.metadata.structures.JClass;
import org.ow2.util.scan.api.metadata.structures.JField;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.rmi.RemoteException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

/**
 * This class manages the injection 
 * @author Florent Benoit
 */
public class ArchiveInjection {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(ArchiveInjection.class);

    /**
     * Defines java.lang.Object class.
     */
    public static final String JAVA_LANG_OBJECT = "java/lang/Object";    

    /**
     * Archive that will be used to analyze the classes.
     */
    private IArchive archive = null;

    /**
     * Analyze of the archive has been done ?
     */
    private boolean analyzed = false;

    /**
     * Remote JNDI Resolver (if present).
     */
    private EZBRemoteJNDIResolver jndiResolver = null;

    /**
     * Collection of metadata that have been analyzed.
     */
    private Collection<? extends ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata>> metadataCollection = null;

    /**
     * Default constructor.
     * @param archive the given archive on which injection is based.
     */
    public ArchiveInjection(final IArchive archive) {
        this.archive = archive;
    }

    /**
     * Analyze the archive.
     * @throws ArchiveInjectionException if analyze fails
     */
    public void analyze() throws ArchiveInjectionException {

        // Get Deployable
        IDeployable<?> deployable = null;
        try {
            deployable = DeployableHelper.getDeployable(archive);
        } catch (DeployableHelperException e) {
            throw new ArchiveInjectionException("Unable to get a deployable on archive '" + archive + "'.", e);
        }



        // Client Archive ?
        if (CARDeployable.class.isInstance(deployable)) {
            // Create metadata
            ICarDeployableMetadata carDeployableMetadata = null;
            try {
                carDeployableMetadata = new CarDeployableMetadataFactory().createDeployableMetadata(CARDeployable.class.cast(deployable));
            } catch (DeployableMetadataException e) {
                throw new ArchiveInjectionException("Unable to get metadata on archive '" + archive + "'.", e);
            }

            // Use only common part
            this.metadataCollection = carDeployableMetadata.getCarClassMetadataCollection();
        }  else {
            //TODO : adapt for other archives
            throw new UnsupportedOperationException("Can only manage injection for Client Deployable. Deployable found is '" + deployable + "'.");
        }

        // Ok, now that metadata are here, 


        analyzed = true;
    }


    /**
     * Initialize stuff in the given class.
     * @param clazz the class to initialize
     */
    public void init(final Class clazz) throws ArchiveInjectionException {
        // Analyze the metadata if not done
        if (!analyzed) {
            analyze();
        }

        // Get the metadata of this class
        Class classToAnalyze = clazz;
        ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> classMetadata;
        classMetadata = getClassMetadata(classToAnalyze);

        // Found metadata ?
        while (classMetadata != null) {

            // Perform injection
            inject(classMetadata, classToAnalyze, null);

            // Super class present ?
            String superClassName = classMetadata.getJClass().getSuperName();
            // If there is a super class, try to find data on it
            if (!JAVA_LANG_OBJECT.equals(superClassName)) {
                String superClassNameStr = superClassName.replace('/', '.');
                classMetadata = getClassMetadata(superClassNameStr);
                // Load the super class
                try {
                    classToAnalyze = classToAnalyze.getClassLoader().loadClass(superClassNameStr);
                } catch (ClassNotFoundException e) {
                    throw new ArchiveInjectionException("Cannot load class '" + superClassNameStr + "'." , e);
                }
            } else {
                // break the loop
                classMetadata = null;
            }

        }

    }

    /**
     * Perform injection in the given instance (or if null on a static way on the class) with the provided metadata
     * @param commonMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for static case.
     */
    public void inject(ICommonClassMetadata commonMetadata, Class clazz, Object instance) throws ArchiveInjectionException {

        // Loop on the fields
       Collection<? extends ICommonFieldMetadata>fieldCollection = commonMetadata.getStandardFieldMetadataCollection();
       Iterator<? extends ICommonFieldMetadata> itField = fieldCollection.iterator();
        while (itField.hasNext()) {
            ICommonFieldMetadata fieldMetadata = itField.next();

            // Set the interface name by using the descriptor of the field
            Type typeInterface = Type.getType(fieldMetadata.getJField().getDescriptor());
            String interfaceName = typeInterface.getClassName();


            // Inject EJB
            IJEjbEJB jejb = fieldMetadata.getJEjbEJB();
            if (jejb != null) {
                // Update interface name
                jejb.setBeanInterface(interfaceName);
                // inject field
                injectFieldEJB(jejb, fieldMetadata, clazz, instance);
            }

            // Inject Resources
            IJAnnotationResource jAnnotationResource = fieldMetadata.getJAnnotationResource();
            if (jAnnotationResource != null) {
                // Update interface name
                jAnnotationResource.setType(interfaceName);
                // inject field
                injectFieldResource(jAnnotationResource, fieldMetadata, clazz, instance);
            }


        }

        // PostConstruct methods
        List<ICommonMethodMetadata> postConstructMethods = new ArrayList<ICommonMethodMetadata>();

        // Loop on the methods
        Collection<? extends ICommonMethodMetadata> methodCollection = commonMetadata.getStandardMethodMetadataCollection();
        Iterator<? extends ICommonMethodMetadata> itMethod = methodCollection.iterator();
         while (itMethod.hasNext()) {
             ICommonMethodMetadata methodMetadata = itMethod.next();

             // PostConstruct Method ?
             if (methodMetadata.isPostConstruct()) {
                postConstructMethods.add(methodMetadata);
             }

         }

        // Call Post Construct methods
        invokePostConstructMethods(postConstructMethods, clazz, instance);
    }


    /**
     * Call postconstruct methods.
     * @param postConstructMethods the methods to call
     * @param clazz the class to use for getting the methods
     * @param instance the given instance to call
     * @throws ArchiveInjectionException if invocation on the methods is not done
     */
    protected void invokePostConstructMethods(List<ICommonMethodMetadata> postConstructMethods, Class clazz, Object instance) throws ArchiveInjectionException {
        // null list, return
        if (postConstructMethods == null) {
            return;
        }

        // Call each method
        for (ICommonMethodMetadata postConstructMethodMetadata : postConstructMethods) {
            Method postConstructMethod = null;
            try {
                postConstructMethod = clazz.getDeclaredMethod(postConstructMethodMetadata.getJMethod().getName());
            } catch (NoSuchMethodException e) {
                throw new ArchiveInjectionException("Cannot invoke postconstruct method", e);
            }

            boolean accessible = postConstructMethod.isAccessible();
            try {
                postConstructMethod.setAccessible(true);
                postConstructMethod.invoke(instance);
            } catch (IllegalAccessException e) {
                throw new ArchiveInjectionException("Cannot invoke postconstruct method", e);
            } catch (InvocationTargetException e) {
                throw new ArchiveInjectionException("Cannot invoke postconstruct method", e);
            } finally {
                postConstructMethod.setAccessible(accessible);
            }
        }

    }


    /**
     * Perform injection in the given instance (or if null on a static way on the class) with the provided metadata
     * @param ejb the EJB-REF that needs to be injected
     * @param fieldMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for static case.
     */
    public void injectFieldEJB(IJEjbEJB ejb, ICommonFieldMetadata fieldMetadata, Class clazz, Object instance) throws ArchiveInjectionException {
        // get data for the EJB
        String interfaceName = ejb.getBeanInterface();
        String beanName = ejb.getBeanName();
        String mappedName = ejb.getMappedName();

        // Check if there is a value bound in the ENC
        String encName = buildENCName(ejb.getName(), fieldMetadata);
        boolean needToBind = true;
        Object value = null;
        try {
            value = new InitialContext().lookup("java:comp/env/" + encName);
            needToBind = false;
        } catch (NamingException e) {
            logger.debug("Value for ENC '{0}' not bound in the registry", encName);
        }

        // Inject this EJB reference

        // JNDI Name to use if no value yet
        if (value == null) {
            String jndiName = null;

            // Mapped name defined ? use it as JNDI name
            if (mappedName != null) {
                jndiName = mappedName;
            } else {
                // needs to ask the remote side
                jndiName = getRemoteJNDIName(interfaceName, beanName);
            }

            // OK, now that JNDI name is here, get value
            try {
                value = new InitialContext().lookup(jndiName);
            } catch (NamingException e) {
                throw new ArchiveInjectionException("Cannot get object with JNDI Name '" + jndiName + "' for ejb '" + ejb + "'.", e);
            }
        }

        // Set value
        setFieldValue(fieldMetadata, clazz, instance, value);

        // bind in ENC
        if (needToBind) {
            bindValue(encName, value);
        }

    }

    /**
     * Perform injection in the given instance (or if null on a static way on the class) with the provided metadata
     * @param jAnnotationResource the resource-ref that needs to be injected
     * @param fieldMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for static case.
     */
    public void injectFieldResource(IJAnnotationResource jAnnotationResource, ICommonFieldMetadata fieldMetadata, Class clazz, Object instance) throws ArchiveInjectionException {
        // get data for the Resource
        String mappedName = jAnnotationResource.getMappedName();

        // Value that will be injected
        Object value = null;

        // Check if there is a value bound in the ENC
        String encName = buildENCName(jAnnotationResource.getName(), fieldMetadata);
        boolean needToBind = true;
        try {
            value = new InitialContext().lookup("java:comp/env/" + encName);
            needToBind = false;
        } catch (NamingException e) {
            logger.debug("Value for ENC '{0}' not bound in the registry", encName);
        }


        // No value, needs to find it
        if (value == null) {
            if (mappedName == null) {
                logger.warn("Injection of @Resource for {0} is not supported", jAnnotationResource);
                return;
            }

            // OK, now that JNDI name is here, get value
            try {
                value = new InitialContext().lookup(mappedName);
            } catch (NamingException e) {
                throw new ArchiveInjectionException("Cannot get object with JNDI Name '" + mappedName + "' for Resource '" + jAnnotationResource + "'.", e);
            }
        }

        // Set value
        setFieldValue(fieldMetadata, clazz, instance, value);

        // bind in ENC
        if (needToBind) {
            bindValue(encName, value);
        }

    }

    /**
     * Bind the given value in the ENC.
     * @param encName the ENC name
     * @param value the ENC value
     */
    protected void bindValue(final String encName, Object value) {
        // bind in ENC if not already done
        try {
            new InitialContext().bind("java:comp/env/" + encName, value);
        } catch (NamingException e) {
            logger.warn("Unable to bind the value for ENC '{0}'.", encName);
        }
    }

    /**
     * Sets the value on a given field by using given arguments
     * @param fieldMetadata the metadata for the name of the field
     * @param clazz the class of the field
     * @param instance the instance of the class
     * @param value the value to set
     * @throws ArchiveInjectionException if the value cannot be set
     */
    protected void setFieldValue(ICommonFieldMetadata fieldMetadata, Class clazz, Object instance, Object value) throws ArchiveInjectionException {

        // Value is obtained, set it
        // Get Field from the class
        JField jField = fieldMetadata.getJField();
        Field field = null;
        try {
            field = clazz.getDeclaredField(jField.getName());
        } catch (NoSuchFieldException e) {
            throw new ArchiveInjectionException("Cannot get field '" + jField + "' on the class '" + clazz.getName() + "'", e);
        }

        boolean accessible = field.isAccessible();
        try {
            // needs to be allowed to change value of the field
            field.setAccessible(true);

            // Set value
            try {
                field.set(instance, value);
            } catch (IllegalAccessException e) {
                throw new ArchiveInjectionException("Cannot set given value on the field '" + jField + "'.", e);
            }
        } finally {
            // set back accessible attribute
            field.setAccessible(accessible);
        }

    }

    /**
     * Build an ENC name for a given field.
     * @param name the existing name of the ENC specified through annotation/xml
     * @param fieldMetadata the metadata of the field
     * @return the updated enc name if there was none
     */
    protected String buildENCName(final String name, final ICommonFieldMetadata fieldMetadata) {
        String encName = name;
        if (encName == null || "".equals(encName)) {
            encName = fieldMetadata.getClassMetadata().getJClass().getName().replace("/", ".") + "/"
                    + fieldMetadata.getJField().getName();
        }
        return encName;
    }

    /**
     * Gets a JNDI name from the server for a given bean/interface.
     * @param interfaceName the name of the interface
     * @param beanName the name of the bean
     * @return a JNDI name if value has been found, else an exception
     * @throws ArchiveInjectionException if JNDI Name is not found
     */
    protected String getRemoteJNDIName(String interfaceName, String beanName) throws ArchiveInjectionException {

        String jndiName = null;
        // Init resolver if not yet defined
        if (jndiResolver == null) {
            // Get object
            Object o = null;
            try {
                o = new InitialContext().lookup("EZB_Remote_JNDIResolver");
            } catch (NamingException e) {
                // No Remote JNDI resolver, so throws exception
                throw new ArchiveInjectionException("No Remote EJB3 JNDI Resolver found", e);
            }

            // Object is here, so we can cast it
            jndiResolver = (EZBRemoteJNDIResolver) PortableRemoteObject.narrow(o, EZBRemoteJNDIResolver.class);

        }

        // Ask the Resolver with interface name and bean name
        List<EZBJNDIData> jndiDataList = null;
        try {
            jndiDataList = jndiResolver.getNames(interfaceName, beanName);
        } catch (RemoteException re) {
            // No Remote JNDI resolver, so throw first exception
            throw new ArchiveInjectionException("Unable to get JNDI Name for '" + interfaceName + "'/'" + beanName + "'", re);
        }

        // Data is here, check if it is empty or not
        if (jndiDataList.size() == 0) {
            throw new ArchiveInjectionException("Unable to get JNDI Name for '" + interfaceName + "'/'" + beanName + "', no data was found on the remote side.");
        } else if (jndiDataList.size() > 1) {
            // too many entries
            throw new ArchiveInjectionException("Unable to get JNDI Name for '" + interfaceName + "'/'" + beanName + "', too many answers : '" + jndiDataList + "'.");
        }

        // Only one item found, so get JNDI Name from this object
        EZBJNDIData jndiData = jndiDataList.get(0);
        // Update JNDI name
        jndiName = jndiData.getName();
        if (logger.isDebugEnabled()) {
            logger.debug("Found JNDI Name '" + jndiName + "' for '" + interfaceName + "'/'" + beanName + "', too many answers : '" + jndiDataList + "'.");
        }

        return jndiName;
    }



    /**
     * Gets the metadata for a given class
     * @param clazz the class on which we want to get metadata
     * @return the metadata if found else null
     */
    protected ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> getClassMetadata(Class clazz) {
        return getClassMetadata(clazz.getName());
    }

    /**
     * Gets the metadata for a given class
     * @param requestedClassName the name of the class on which we want to get metadata
     * @return the metadata if found else null
     */
    protected ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> getClassMetadata(final String requestedClassName) {

        // Get an iterator
        Iterator<? extends ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata>> it = metadataCollection.iterator();

        // Search the class by iterating on the metadata collection
        while (it.hasNext()) {
            // Declare temporary var
            ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> tmpMetadata;

            // Get current metadata
            tmpMetadata = it.next();

            // Metadata for the expected class ?
            JClass jClass = tmpMetadata.getJClass();
            if (jClass != null) {
                String className = jClass.getName();
                // compare name
                if (requestedClassName.equals(className.replace('/', '.')))  {
                    return tmpMetadata;
                }

            }
        }

        // Not found
        return null;
    }

}
